structure KCikcakk :> KCikcakk =
struct

(* pp00s nhf, v1.2 -- 2000. marc. 25. Hanak Peter *)

(* cikcakk_in f = A f szoveges allomanyban leirt maximalis mezoertek es feladvanymatrix; 
     unix alatt cikcakk_in("/dev/stdin"), MS-DOS alatt cikcakk_in("con")
     a billentyuzetrol olvassa be a feladvanymatrixot.
   cikcakk_in : string -> (int * TCikcakk.field TCikcakk.matrix)
*)
fun cikcakk_in f = 
    let 
	fun getLines istr =
	      if TextIO.endOfStream istr
              then []
	      else List.mapPartial Int.fromString 
		     (String.tokens Char.isSpace (TextIO.inputLine istr)) ::
		   getLines istr

	val istr = TextIO.openIn f
	val mx = valOf(Int.fromString(TextIO.inputLine istr))
	val brd = getLines istr
	val _ = TextIO.closeIn istr
    in
	(mx, brd)
    end;

(*
*)

fun toString TCikcakk.nw = "nw"
  | toString TCikcakk.n	 = "n"
  | toString TCikcakk.ne = "ne"
  | toString TCikcakk.w	 = "w"
  | toString TCikcakk.e	 = "e"
  | toString TCikcakk.sw = "sw"
  | toString TCikcakk.s	 = "s"
  | toString TCikcakk.se = "se"

(* sepLine (ms, pad) = ms mezoszamanak megfelelo hosszusagu,
             csak pad szeles elvalaszto jelekbol allo fuzer.
   sepLine : TCikcakk.field TCikcakk.row * int -> string
*)
fun sepLine (_::ms, pad) = StringCvt.padLeft #"-" (pad+1) "+" ^ sepLine (ms, pad)
  | sepLine ([], _) = "";

(* valLine (fs, ms, pad) = ms mezoszamanak megfelelo hosszusagu,
       mezoertekekbol es elvalaszto jelekbol allo fuzer
       (fs es ms hossza egyforma), a mezoszelesseg pad
   valLine : TCikcakk.field TCikcakk.row * TCikcakk.dir TCikcakk.row * int -> string
*)
fun valLine (f::fs, m::ms, pad) = StringCvt.padLeft #" " pad
                                             (toString m ^ Int.toString f) ^
                             "|" ^ valLine(fs, ms, pad)
  | valLine _ = "";

(* max_elem xss m = az xss egeszek maximalis elemenek es m egesznek a
                    maximuma
   max_elem : int -> int list list -> int
*)
val max_elem = foldl ((fn f => fn (a, b) => f b a) (foldl Int.max))

(* drawBoard (fss, mss) = Az mss megoldasmatrix altal lefedett fss
      feladvanymatrixot adja eredmenyul.
   drawboard : TCikcakk.field TCikcakk.matrix * TCikcakk.dir TCikcakk.matrix -> string
*)
fun drawBoard (fss, mss) =
      let
	  val pad = floor (Math.log10 (Real.fromInt (max_elem 1 fss))) + 3
          fun dB ([fs], [ms]) = "+" ^ sepLine (fs, pad) ^ "\n" ^
                                "|" ^ valLine (fs, ms, pad) ^ "\n" ^
                                "+" ^ sepLine (fs, pad) ^ "\n"
            | dB (fs::fss, ms::mss) = "+" ^ sepLine (fs, pad) ^ "\n" ^
                                      "|" ^ valLine (fs, ms, pad) ^ "\n" ^
                                      dB (fss, mss)
            | dB _ = "";
      in
          dB (fss, mss)
      end

(* cikcakk_out(m, fss, mss) = Az mss megoldasmatrix altal lefedett fss
       feladvanymatrixot kiirja az m szoveges allomanyba;
     unix alatt cikcakk_out("/dev/stdout", fss, mss), MS-DOS alatt
     cikcakk_out("con", fss, mss) a kepernyore irja ki a megoldasmatrixot.
   cikcakk_out : string * TCikcakk.field TCikcakk.matrix * 
                         TCikcakk.dir TCikcakk.matrix -> unit
*)
fun cikcakk_out (m, fss, mss) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, drawBoard (fss, mss))
      in
	  TextIO.closeOut ostr
      end;

(* solve(f, m) = Beolvas egy feladvanyt az f szoveges allomanybol, amelynek
                 elso soraban a maximalis mezoertek van, es kiirja
                 az osszes megoldast az m szoveges allomanyba; ehhez
                 termeszetesen szuksege van a Cikcakk.cikcakk fuggvenyre.
                 Eredmenye a feladvanymatrix nevet, a megoldasmatrix
                 meretet es a futasi idot mp-ben tartalmazo fuzer.
   solve : string * string -> string
*)
fun solve (f, m) =
    let val (mx, board) = cikcakk_in f
	val ostr = TextIO.openOut m
	fun putBoard sol = TextIO.output(ostr, drawBoard(board, sol) ^ "\n")
	val starttime = Timer.startCPUTimer()
	val mboard = Cikcakk.cikcakk board mx
	val {usr=tim,...} = Timer.checkCPUTimer starttime
	val _ = map putBoard mboard
	val _ = TextIO.closeOut ostr
	val t = "Table: " ^ f ^ ", " ^
                Int.toString(length mboard) ^ " solution(s), " ^
		"time: " ^ Time.fmt 2 tim ^ " sec\n"
    in
	t
    end;

end
